"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var metadata_1 = require("../../server/analysis/metadata");
/**
 * Generates API block from the component type.
 * @param component Component type.
 */
function astApi(component) {
    var meta = metadata_1.asMeta(component);
    if (!meta) {
        return "<p>⚠️ No API docs available. Try to rebuild the story module.</p>";
    }
    return ("<h2 id=\"API\">API " + (meta.gitUrl
        ? "<a class=\"n9-h-l\" href=\"" + meta.gitUrl + "\" target=\"_blank\">view source &lt;/&gt;</a>"
        : "") + "</h2>\n    " + (meta.jsDoc ? "<div class=\"sb-jsdoc\">" + meta.jsDoc + "</div>" : "") +
        astUsage(component) +
        renderApiProps(meta.inputs, "Inputs") +
        renderApiProps(meta.outputs, "Outputs") +
        renderApiProps(meta.nonAngular.props, "Properties") +
        renderApiMethods(meta.methods));
}
exports.astApi = astApi;
/**
 * Generates preview template from the component type.
 * @param component Component type.
 * @param transclude Transclude template inside `<ng-content>`.
 * @param noWrap Disable component wrapping.
 */
function astPreview(component, transclude, noWrap) {
    if (noWrap === void 0) { noWrap = false; }
    var meta = metadata_1.asMeta(component);
    var annotated = metadata_1.asAnnotated(component);
    if (!meta || !annotated) {
        return "<p>⚠️ No preview available. Try to rebuild the story module.</p>";
    }
    return noWrap
        ? getTemplateForType(meta, annotated, transclude)
        : wrapPreview(getTemplateForType(meta, annotated, transclude), meta.kind !== metadata_1.TypeKind.Pipe, false);
}
exports.astPreview = astPreview;
/**
 * Generates usage block i.e. `<selector [...inputs] (...outputs)>`.
 * @param component Component type.
 */
function astUsage(component) {
    var meta = metadata_1.asMeta(component);
    var annotated = metadata_1.asAnnotated(component);
    if (!meta || !annotated) {
        return "<p>⚠️ No usage available. Try to rebuild the story module.</p>";
    }
    switch (meta.kind) {
        case metadata_1.TypeKind.Component: {
            return "<pre><code class=\"language-html\">" + getComponentUsage(meta, annotated.selector) + "</code></pre>";
        }
        default:
            return "";
    }
}
exports.astUsage = astUsage;
function wrapPreview(preview, showRuler, noBg) {
    return (preview &&
        "<div class=\"component-preview" + (showRuler ? " w-ruler" : "") + "\">\n      " + (noBg ? preview : "<div class=\"component-preview-bg\">" + preview + "</div>") + "\n    </div>");
}
exports.wrapPreview = wrapPreview;
function getTemplateForType(meta, annotation, transclude) {
    if (!meta) {
        return "";
    }
    if (meta.kind === metadata_1.TypeKind.Component && annotation) {
        var selector = annotation.selector;
        return "<" + selector + " #preview " + generateBindingTemplate(meta) + ">" + (transclude ||
            "") + "</" + selector + ">";
    }
    if (meta.kind === metadata_1.TypeKind.Pipe) {
        return "<n9-pipe-usage></n9-pipe-usage>";
    }
    return "";
}
function generateBindingTemplate(metadata) {
    return metadata.inputs
        .map(function (p) { return "[" + p.name + "]=\"" + p.name + "\""; })
        .concat(metadata.outputs.filter(function (p) { return p.defaultValue; }).map(function (p) { return "(" + p.name + ")=\"" + p.name + "($event)\""; }))
        .join(" ");
}
function getComponentUsage(meta, selectorMeta) {
    var lt = "<span class=\"n9-h-m\">&lt;</span>";
    var ltc = "<span class=\"n9-h-m\">&lt;/</span>";
    var gt = "<span class=\"n9-h-m\">&gt;</span>";
    var props = meta.inputs
        .map(function (p) { return "<span class=\"n9-h-m\">[" + p.name + "]</span><span class=\"n9-h-v\">=\" \"</span>"; })
        .concat(meta.outputs.map(function (p) { return "<span class=\"n9-h-m\">(" + p.name + ")</span><span class=\"n9-h-v\">=\" \"</span>"; }));
    var propsToken = props.length ? " " + props.join("\r\n\t") : "";
    var selector = "<span class=\"n9-h-t\">" + selectorMeta + "</span>";
    return /[a-zA-Z0-9\-]+$/.test(selectorMeta)
        ? "" + lt + selector + propsToken + gt + "\r\n" + ltc + selector + gt
        : selector;
}
function renderApiProps(props, title) {
    if (!props.length) {
        return "";
    }
    var propTpls = props
        .map(function (p) {
        return "<div class=\"n9-p-c\">\n      " + (p.jsDoc ? "<div class=\"sb-jsdoc\">" + p.jsDoc + "</div>" : "") + "\n\n      " + (p.kind !== metadata_1.PropKind.Regular ? "<span class=\"n9-p-m\">" + getKind(p.kind) + "</span>" : "") + "\n\n      <span class=\"n9-p-n\">" + p.name + (p.type ? ":</span> <span class=\"n9-p-t\">" + p.type + "</span>" : "") + "\n      \n      " + (p.defaultValue ? "<span class=\"n9-p-d\"> = " + p.defaultValue + "</span>" : "") + "\n    </div>";
    })
        .join("");
    return "<h3 id=\"" + title + "\">" + title + "</h3>" + propTpls;
}
function getKind(kind) {
    if (kind === metadata_1.PropKind.OnlyGetter) {
        return "get";
    }
    return "set";
}
function renderApiMethods(methods) {
    if (!methods.length) {
        return "";
    }
    var fnTpl = function (m) {
        return "<span class=\"n9-p-n\">" +
            ("<span class=\"n9-m-n\">" + m.name + "</span>(") +
            m.args
                .map(function (arg, i) {
                return "<span class=\"n9-p-n\">" + arg.name + "</span>" + (arg.type ? ": <span class=\"n9-p-t n9-p-c\">" + arg.type + "</span>" : "") +
                    ("" + ((arg.defaultValue && "<span class=\"n9-p-d\"> = " + arg.defaultValue + "</span>") || "")) +
                    ("" + (i === m.args.length - 1 ? "" : ", "));
            })
                .join("") +
            (")</span>" + (m.returnType ? ": <span class=\"n9-p-t\">" + m.returnType + "</span>" : ""));
    };
    return ("<h3 id=\"methods\">Methods</h3>" +
        methods
            .map(function (m) { return "<div *ngFor=\"let m of methods\" class=\"n9-p-c\">\n       " + (m.jsDoc ? "<div class=\"sb-jsdoc\">" + m.jsDoc + "</div>" : "") + " " + fnTpl(m) + "\n      </div>"; })
            .join(""));
}
