import { NgModule, Type } from "@angular/core";
export interface StoryTemplate {
    beforePreview?: string;
    preview?: string;
    afterPreview?: string;
}
export interface Story<T> {
    component?: Type<T>;
    template?: string | StoryTemplate;
    styles?: string[];
    moduleMetadata?: NgModule;
    props?: ComponentProps<T>;
}
export declare type StoryFn<T> = () => Story<T>;
export interface StoryApi {
    /**
     * Adds the new story.
     * @param name Story name.
     * @param storyFn Story factory.
     * @param parameters Story parameters.
     */
    add<T>(name: string, storyFn: StoryFn<T>, parameters?: StoryParameters): StoryApi;
    /**
     * Generates the story with preview, usage and API blocks.
     * @param name Story name.
     * @param component Component type.
     * @param config Docs config.
     * @param parameters Story parameters.
     */
    addApiDocs<T>(name: string, component: Type<T>, config?: () => ApiDocsConfig<T>, parameters?: StoryParameters): StoryApi;
    /**
     * Generates the story with pure markdown template. Angular templates don't work here.
     * @param name Story name.
     * @param markdown Markdown template.
     */
    addWikiDocs(name: string, markdown: string): StoryApi;
    /**
     * Adds decorator to the stories.
     * @param decorator Decorator fn.
     */
    addDecorator(decorator: any): StoryApi;
}
/**
 * Markdown section.
 */
export interface MarkdownSection {
    /**
     * Section key. In `*.md` file defined as: `<!-- key -->`.
     */
    key: string;
    /**
     * Template string for the section.
     */
    template: string;
}
/**
 * API docs configuration.
 */
export interface ApiDocsConfig<T> {
    /**
     * @NgModule metadata for the story. Note: declararion for the component is included by default, but
     * excluded if custom declaration was provided.
     */
    moduleMetadata?: NgModule;
    /**
     * Manual props - these will override any auto-generated props.
     */
    props?: ComponentProps<T>;
    /**
     * API docs template can be used to provide custom template.
     */
    template?: ApiDocsTemplate;
    /**
     * Styles for the API docs.
     */
    styles?: string[];
    /**
     * Background settings for preview.
     */
    background?: "none";
}
/**
 * Component properties.
 */
export declare type ComponentProps<T> = {
    [P in keyof T]?: any;
} | {
    [propName: string]: any;
};
export interface StoryParameters {
    [name: string]: any;
}
/**
 * API docs template sections.
 */
export interface ApiDocsTemplate {
    /**
     * Header template (comes before `beforeApiDocs`).
     */
    header?: string;
    /**
     * Template for component preview. Note: replaces auto-generated preview.
     */
    preview?: string | false;
    /**
     * Template for component content in preview.
     */
    previewTransclude?: string;
    /**
     * Template before API section.
     */
    beforeApiDocs?: string;
    /**
     * API section markup. Note: replaces auto-generated markup.
     */
    apiDocs?: string | false;
    /**
     * Template after API section.
     */
    afterApiDocs?: string;
}
