"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
Object.defineProperty(exports, "__esModule", { value: true });
var ts_simple_ast_1 = require("ts-simple-ast");
var path = require("path");
var marked = require("marked");
var metadata_1 = require("./metadata");
var ANALYSIS_TRIGGERS = {
    astProps: 0,
    astPreview: 0,
    astUsage: 0,
    astApi: 0,
    addApiDocs: 1
};
function findTriggers(sourceFile, reader) {
    var argCache = {};
    return sourceFile
        .getDescendantsOfKind(ts_simple_ast_1.SyntaxKind.CallExpression)
        .map(function (ce) {
        var expr = ce.getExpression();
        var fnName = expr.getName ? expr.getName() : expr.getText();
        return { ce: ce, ceIdx: ANALYSIS_TRIGGERS[fnName] };
    })
        .filter(function (_a) {
        var ceIdx = _a.ceIdx;
        return ceIdx != null;
    })
        .map(function (_a) {
        var ce = _a.ce, ceIdx = _a.ceIdx;
        var targetArg = ce.getArguments()[ceIdx];
        var cacheKey = targetArg.getText();
        if (argCache[cacheKey]) {
            return {
                replacement: {
                    start: targetArg.getStart(),
                    end: targetArg.getEnd(),
                    text: argCache[cacheKey]
                }
            };
        }
        var declaration = findDeclaration(targetArg.getText(), sourceFile, reader);
        if (!declaration) {
            return null;
        }
        var declarationKey = declaration.getSourceFile().getFilePath() + "+" + cacheKey;
        return {
            key: declarationKey,
            filePath: declaration.getSourceFile().getFilePath(),
            declaration: declaration,
            replacement: {
                start: targetArg.getStart(),
                end: targetArg.getEnd(),
                text: (argCache[cacheKey] = "{k:" + JSON.stringify(declarationKey) + ",c:" + targetArg.getText() + "} as any")
            }
        };
    })
        .filter(function (r) { return r; });
}
exports.findTriggers = findTriggers;
function findDeclaration(accessExpression, sourceFile, reader) {
    var argElements = accessExpression.split(".");
    var complexModule = argElements.length > 1;
    var className = complexModule ? argElements[argElements.length - 1] : argElements[0];
    var importStatement = sourceFile.getImportDeclaration(function (n) {
        var named = n.getNamedImports().some(function (x) {
            var nameFound = x.getName() === className;
            if (nameFound) {
                return true;
            }
            var alias = x.getAliasIdentifier();
            if (alias && alias.getText() === className) {
                className = x.getName();
                return true;
            }
        });
        if (!complexModule || named) {
            return named;
        }
        var namespace = n.getFirstDescendantByKind(ts_simple_ast_1.SyntaxKind.NamespaceImport);
        if (namespace) {
            var identifier = namespace.getFirstDescendantByKind(ts_simple_ast_1.SyntaxKind.Identifier);
            if (identifier) {
                return identifier.getText() === argElements[0];
            }
        }
        return false;
    });
    if (!importStatement) {
        return sourceFile.getClass(className);
    }
    var _a = findClassInRelativePath(sourceFile.getFilePath(), importStatement.getModuleSpecifierValue(), className, reader), declaration = _a.declaration, targetSource = _a.source;
    if (declaration) {
        return declaration;
    }
    if (!targetSource) {
        return null;
    }
    // try search in barrel exports
    var exportStatement = targetSource.getExportDeclaration(function (n) {
        return n.getNamedExports().some(function (x) { return x.getNameNode().getText() === className; });
    });
    if (!exportStatement) {
        return null;
    }
    return findClassInRelativePath(targetSource.getFilePath(), exportStatement.getModuleSpecifierValue(), className, reader).declaration;
}
function findClassInRelativePath(base, relative, className, reader) {
    var targetPath = path.resolve(path.dirname(base), relative) + ".ts";
    var fileContents = reader(targetPath);
    if (!fileContents) {
        return {};
    }
    var targetSource = ts_simple_ast_1.createWrappedNode(ts_simple_ast_1.ts.createSourceFile(targetPath, fileContents, ts_simple_ast_1.ScriptTarget.Latest, true));
    return {
        declaration: targetSource.getClass(className),
        source: targetSource
    };
}
function getMetadata(declaration, markedOptions, reader) {
    var multiMetadata = [];
    var currentClass = declaration;
    while (currentClass) {
        multiMetadata.unshift(getSingleClassMetadata(currentClass, markedOptions));
        var baseClass = currentClass.getExtends();
        if (!baseClass) {
            break;
        }
        currentClass = findDeclaration(baseClass.getExpression().getText(), currentClass.getSourceFile(), reader);
    }
    return multiMetadata.reduce(function (p, n) {
        return {
            kind: n.kind,
            inputs: mergeNameable(p.inputs, n.inputs),
            outputs: mergeNameable(p.outputs, n.outputs),
            nonAngular: {
                props: mergeNameable(p.nonAngular.props, n.nonAngular.props)
            },
            methods: mergeNameable(p.methods, n.methods),
            jsDoc: n.jsDoc
        };
    });
}
exports.getMetadata = getMetadata;
function mergeNameable(prev, next) {
    return prev.filter(function (p) { return !next.some(function (n) { return n.name === p.name; }); }).concat(next);
}
function getSingleClassMetadata(declaration, markedOptions) {
    var metadata = {
        kind: analyzeKind(declaration),
        inputs: [],
        outputs: [],
        nonAngular: {
            props: []
        },
        methods: [],
        jsDoc: getJSDoc(declaration, markedOptions, true)
    };
    analyzeProps(declaration, metadata, markedOptions);
    analyzeMethods(declaration, metadata, markedOptions);
    return metadata;
}
function analyzeKind(declaration) {
    for (var _i = 0, _a = declaration.getDecorators(); _i < _a.length; _i++) {
        var decorator = _a[_i];
        switch (decorator.getName()) {
            case "Pipe":
                return metadata_1.TypeKind.Pipe;
            case "Directive":
                return metadata_1.TypeKind.Directive;
            case "Component":
                return metadata_1.TypeKind.Component;
            default:
                break;
        }
    }
    return metadata_1.TypeKind.NonAngular;
}
function analyzeProps(declaration, metadata, markedOptions) {
    var props = declaration.getProperties();
    var propCache = {};
    for (var _i = 0, props_1 = props; _i < props_1.length; _i++) {
        var prop = props_1[_i];
        if (prop.hasModifier(ts_simple_ast_1.SyntaxKind.PrivateKeyword)) {
            continue;
        }
        var propName = prop.getName();
        analyzeDecorators(prop, metadata, {
            prop: {
                name: propName,
                type: getType(propName, prop, propCache),
                kind: metadata_1.PropKind.Regular,
                defaultValue: getDefaultValue(prop),
                jsDoc: getJSDoc(prop, markedOptions)
            }
        });
    }
    for (var _a = 0, _b = declaration.getSetAccessors(); _a < _b.length; _a++) {
        var setter = _b[_a];
        if (setter.hasModifier(ts_simple_ast_1.SyntaxKind.PrivateKeyword)) {
            continue;
        }
        var setterName = setter.getName();
        analyzeDecorators(setter, metadata, {
            prop: {
                name: setterName,
                type: getType(setterName, setter.getParameters()[0], propCache),
                kind: metadata_1.PropKind.OnlySetter,
                jsDoc: getJSDoc(setter, markedOptions)
            }
        });
    }
    for (var _c = 0, _d = declaration.getGetAccessors(); _c < _d.length; _c++) {
        var getter = _d[_c];
        if (getter.hasModifier(ts_simple_ast_1.SyntaxKind.PrivateKeyword)) {
            continue;
        }
        var getterName = getter.getName();
        analyzeDecorators(getter, metadata, {
            prop: {
                name: getterName,
                type: getType(getterName, getter, propCache),
                kind: metadata_1.PropKind.OnlyGetter,
                jsDoc: getJSDoc(getter, markedOptions)
            }
        });
    }
}
function analyzeDecorators(node, metadata, nodeMeta) {
    if (nodeMeta.prop && nodeMeta.prop.jsDoc === false) {
        return;
    }
    for (var _i = 0, _a = node.getDecorators(); _i < _a.length; _i++) {
        var decorator = _a[_i];
        switch (decorator.getName()) {
            case "Input":
                nodeMeta.prop &&
                    metadata.inputs.push(__assign({}, nodeMeta.prop, { name: tryGetAliasedPropName(decorator) || nodeMeta.prop.name }));
                return;
            case "Output":
                nodeMeta.prop &&
                    metadata.outputs.push(__assign({}, nodeMeta.prop, { name: tryGetAliasedPropName(decorator) || nodeMeta.prop.name }));
                return;
            default:
                break;
        }
    }
    nodeMeta.prop && metadata.nonAngular.props.push(nodeMeta.prop);
}
function tryGetAliasedPropName(decorator) {
    var nameArg = decorator.getArguments()[0];
    try {
        return nameArg ? eval(nameArg.getText()).toString() : null;
    }
    catch (_a) {
        return null;
    }
}
function analyzeMethods(declaration, metadata, markedOptions) {
    var methods = declaration.getMethods();
    for (var _i = 0, methods_1 = methods; _i < methods_1.length; _i++) {
        var method = methods_1[_i];
        if (method.hasModifier(ts_simple_ast_1.SyntaxKind.PrivateKeyword)) {
            continue;
        }
        var fnDef = {
            name: method.getName(),
            args: method.getParameters().map(function (p) {
                return ({
                    name: p.getName(),
                    type: getType(p.getName(), p),
                    defaultValue: getDefaultValue(p)
                });
            }),
            returnType: getType(method.getName(), method),
            jsDoc: getJSDoc(method, markedOptions)
        };
        if (fnDef.jsDoc === false) {
            continue;
        }
        metadata.methods.push(fnDef);
    }
}
function getDefaultValue(expression) {
    var initializer = expression.getInitializer();
    return initializer && initializer.getText();
}
function getJSDoc(node, markedOptions, forceReturn) {
    if (forceReturn === void 0) { forceReturn = false; }
    var docs = node.getJsDocs();
    var docsText = docs.map(function (d) { return d.getInnerText(); }).join("");
    if (!docsText) {
        return null;
    }
    if (!forceReturn && docs.some(function (d) { return d.getTags().some(function (tag) { return tag.getName() === "internal"; }); })) {
        return false;
    }
    return marked(docsText, markedOptions);
}
function getType(name, node, propTypeCache) {
    if (propTypeCache === void 0) { propTypeCache = {}; }
    var type = getTypeInner(node, propTypeCache);
    propTypeCache[name] = type;
    return type;
}
function getTypeInner(node, propTypeCache) {
    var simpleTypeNode = (node.getTypeNode && node.getTypeNode()) ||
        (node.getReturnTypeNode && node.getReturnTypeNode());
    if (simpleTypeNode) {
        return simpleTypeNode.getText();
    }
    var initializer = node.getInitializer && node.getInitializer();
    if (initializer) {
        var initializerType = tryGetInitializer(initializer, propTypeCache);
        if (initializerType) {
            return initializerType;
        }
    }
    var fnBody = node.getBody && node.getBody();
    if (fnBody) {
        var fnReturnType = tryGetFnReturnType(fnBody, propTypeCache);
        if (fnReturnType) {
            return fnReturnType;
        }
    }
    return "";
}
function tryGetInitializer(initializer, propTypeCache) {
    if (initializer.getLiteralValue) {
        return typeof initializer.getLiteralValue();
    }
    switch (initializer.getKind()) {
        case ts_simple_ast_1.SyntaxKind.NewExpression:
            var typeArgs = initializer
                .getTypeArguments()
                .map(function (a) { return a.getText(); })
                .join(", ");
            return initializer.getExpression().getText() + (typeArgs ? "<" + typeArgs + ">" : "");
        case ts_simple_ast_1.SyntaxKind.ObjectLiteralExpression:
            return initializer.getText();
        case ts_simple_ast_1.SyntaxKind.PropertyAccessExpression:
            var cachedType = propTypeCache[initializer.getName()];
            if (cachedType) {
                return cachedType;
            }
            break;
        case ts_simple_ast_1.SyntaxKind.ArrowFunction:
            return "Function";
        default:
            break;
    }
    try {
        return typeof eval(initializer.getText());
    }
    catch (_a) { }
}
function tryGetFnReturnType(fnBody, propTypeCache) {
    var returnStatements = fnBody
        .getDescendantsOfKind(ts_simple_ast_1.SyntaxKind.ReturnStatement)
        .filter(function (r) { return r.getExpression(); });
    if (!returnStatements.length) {
        return "void";
    }
    return returnStatements
        .map(function (r) { return tryGetInitializer(r.getExpression(), propTypeCache); })
        .find(function (t) { return !!t; });
}
